@@ -23,52 +23,59 @@
 
 #include <glib.h>
 #include <gdbus.h>
+#include <gw-obex.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/sdp.h>
-#include <gw-obex.h>
+
+struct agent_data;
+struct session_callback;
 
 struct session_data {
 	gint refcount;
 	bdaddr_t src;
 	bdaddr_t dst;
 	uint8_t channel;
+	char *service;		/* Service friendly name */
 	const char *target;	/* OBEX Target UUID */
 	int target_len;
 	uuid_t uuid;		/* Bluetooth Service Class */
-	gchar *name;
 	gchar *path;		/* Session path */
-	gchar *transfer_path;	/* Transfer path */
-	int sock;
-	int fd;
 	DBusConnection *conn;
 	DBusMessage *msg;
 	GwObex *obex;
-	GwObexXfer *xfer;
-	char *buffer;
-	size_t buffer_len;
-	int filled;
-	gint64 size;
-	gint64 transferred;
-	gchar *filename;
-	gchar *agent_name;
-	gchar *agent_path;
-	guint agent_watch;
+	GIOChannel *io;
+	struct agent_data *agent;
+	struct session_callback *callback;
 	gchar *owner;		/* Session owner */
-	guint owner_watch;
-	GPtrArray *pending;
+	guint watch;
+	GSList *pending;
 	void *priv;
 };
 
 typedef void (*session_callback_t) (struct session_data *session,
-							void *user_data);
+					GError *err, void *user_data);
+
+struct session_data *session_create(const char *source,
+						const char *destination,
+						const char *service,
+						uint8_t channel,
+						const char *owner,
+						session_callback_t function,
+						void *user_data);
+
+struct session_data *session_ref(struct session_data *session);
+void session_unref(struct session_data *session);
+void session_shutdown(struct session_data *session);
+
+int session_set_owner(struct session_data *session, const char *name,
+			GDBusWatchFunction func);
+const char *session_get_owner(struct session_data *session);
 
-int session_create(const char *source,
-			const char *destination, const char *target,
-			uint8_t channel, session_callback_t function,
-			void *user_data);
 int session_set_agent(struct session_data *session, const char *name,
 							const char *path);
+const char *session_get_agent(struct session_data *session);
+
 int session_send(struct session_data *session, const char *filename,
 				const char *remotename);
 int session_get(struct session_data *session, const char *type,
