@@ -42,33 +42,185 @@
 
 #include <openobex/obex.h>
 
-#include "logging.h"
+#include "log.h"
 #include "obex.h"
 #include "obex-priv.h"
+#include "server.h"
 #include "dbus.h"
 #include "mimetype.h"
 #include "service.h"
+#include "transport.h"
 #include "btio.h"
 
 /* Default MTU's */
 #define DEFAULT_RX_MTU 32767
 #define DEFAULT_TX_MTU 32767
 
+/* Challenge request */
+#define NONCE_TAG 0x00
+#define OPTIONS_TAG 0x01 /* Optional */
+#define REALM_TAG 0x02 /* Optional */
+
+#define NONCE_LEN 16
+
+/* Challenge response */
+#define DIGEST_TAG 0x00
+#define USER_ID_TAG 0x01 /* Optional */
+#define DIGEST_NONCE_TAG 0x02 /* Optional */
+
 /* Connection ID */
-static guint32 cid = 0x0000;
+static uint32_t cid = 0x0000;
 
 static GSList *sessions = NULL;
 
 typedef struct {
-	guint8  version;
-	guint8  flags;
-	guint16 mtu;
+	uint8_t  version;
+	uint8_t  flags;
+	uint16_t mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
+struct auth_header {
+	uint8_t tag;
+	uint8_t len;
+	uint8_t val[0];
+} __attribute__ ((packed));
+
+static struct {
+	int evt;
+	const char *name;
+} obex_event[] = {
+	/* Progress has been made */
+	{ OBEX_EV_PROGRESS,	"PROGRESS"	},
+	/* An incoming request is about to come */
+	{ OBEX_EV_REQHINT,	"REQHINT"	},
+	/* An incoming request has arrived */
+	{ OBEX_EV_REQ,		"REQ"		},
+	/* Request has finished */
+	{ OBEX_EV_REQDONE,	"REQDONE"	},
+	/* Link has been disconnected */
+	{ OBEX_EV_LINKERR,	"LINKERR"	},
+	/* Malformed data encountered */
+	{ OBEX_EV_PARSEERR,	"PARSEERR"	},
+	/* Connection accepted */
+	{ OBEX_EV_ACCEPTHINT,	"ACCEPTHINT"	},
+	/* Request was aborted */
+	{ OBEX_EV_ABORT,	"ABORT"		},
+	/* Need to feed more data when sending a stream */
+	{ OBEX_EV_STREAMEMPTY,	"STREAMEMPTY"	},
+	/* Time to pick up data when receiving a stream */
+	{ OBEX_EV_STREAMAVAIL,	"STREAMAVAIL"	},
+	/* Unexpected data, not fatal */
+	{ OBEX_EV_UNEXPECTED,	"UNEXPECTED"	},
+	/* First packet of an incoming request has been parsed */
+	{ OBEX_EV_REQCHECK,	"REQCHECK"	},
+	{ 0xFF,			NULL		},
+};
+
+/* Possible commands */
+static struct {
+	int cmd;
+	const char *name;
+} obex_command[] = {
+	{ OBEX_CMD_CONNECT,	"CONNECT"	},
+	{ OBEX_CMD_DISCONNECT,	"DISCONNECT"	},
+	{ OBEX_CMD_PUT,		"PUT"		},
+	{ OBEX_CMD_GET,		"GET"		},
+	{ OBEX_CMD_SETPATH,	"SETPATH"	},
+	{ OBEX_CMD_SESSION,	"SESSION"	},
+	{ OBEX_CMD_ABORT,	"ABORT"		},
+	{ OBEX_FINAL,		"FINAL"		},
+	{ 0xFF,			NULL		},
+};
+
+/* Possible Response */
+static struct {
+	int rsp;
+	const char *name;
+} obex_response[] = {
+	{ OBEX_RSP_CONTINUE,			"CONTINUE"		},
+	{ OBEX_RSP_SWITCH_PRO,			"SWITCH_PRO"		},
+	{ OBEX_RSP_SUCCESS,			"SUCCESS"		},
+	{ OBEX_RSP_CREATED,			"CREATED"		},
+	{ OBEX_RSP_ACCEPTED,			"ACCEPTED"		},
+	{ OBEX_RSP_NON_AUTHORITATIVE,		"NON_AUTHORITATIVE"	},
+	{ OBEX_RSP_NO_CONTENT,			"NO_CONTENT"		},
+	{ OBEX_RSP_RESET_CONTENT,		"RESET_CONTENT"		},
+	{ OBEX_RSP_PARTIAL_CONTENT,		"PARTIAL_CONTENT"	},
+	{ OBEX_RSP_MULTIPLE_CHOICES,		"MULTIPLE_CHOICES"	},
+	{ OBEX_RSP_MOVED_PERMANENTLY,		"MOVED_PERMANENTLY"	},
+	{ OBEX_RSP_MOVED_TEMPORARILY,		"MOVED_TEMPORARILY"	},
+	{ OBEX_RSP_SEE_OTHER,			"SEE_OTHER"		},
+	{ OBEX_RSP_NOT_MODIFIED,		"NOT_MODIFIED"		},
+	{ OBEX_RSP_USE_PROXY,			"USE_PROXY"		},
+	{ OBEX_RSP_BAD_REQUEST,			"BAD_REQUEST"		},
+	{ OBEX_RSP_UNAUTHORIZED,		"UNAUTHORIZED"		},
+	{ OBEX_RSP_PAYMENT_REQUIRED,		"PAYMENT_REQUIRED"	},
+	{ OBEX_RSP_FORBIDDEN,			"FORBIDDEN"		},
+	{ OBEX_RSP_NOT_FOUND,			"NOT_FOUND"		},
+	{ OBEX_RSP_METHOD_NOT_ALLOWED,		"METHOD_NOT_ALLOWED"	},
+	{ OBEX_RSP_NOT_ACCEPTABLE,		"NOT_ACCEPTABLE"	},
+	{ OBEX_RSP_PROXY_AUTH_REQUIRED,		"PROXY_AUTH_REQUIRED"	},
+	{ OBEX_RSP_REQUEST_TIME_OUT,		"REQUEST_TIME_OUT"	},
+	{ OBEX_RSP_CONFLICT,			"CONFLICT"		},
+	{ OBEX_RSP_GONE,			"GONE"			},
+	{ OBEX_RSP_LENGTH_REQUIRED,		"LENGTH_REQUIRED"	},
+	{ OBEX_RSP_PRECONDITION_FAILED,		"PRECONDITION_FAILED"	},
+	{ OBEX_RSP_REQ_ENTITY_TOO_LARGE,	"REQ_ENTITY_TOO_LARGE"	},
+	{ OBEX_RSP_REQ_URL_TOO_LARGE,		"REQ_URL_TOO_LARGE"	},
+	{ OBEX_RSP_UNSUPPORTED_MEDIA_TYPE,	"UNSUPPORTED_MEDIA_TYPE"},
+	{ OBEX_RSP_INTERNAL_SERVER_ERROR,	"INTERNAL_SERVER_ERROR"	},
+	{ OBEX_RSP_NOT_IMPLEMENTED,		"NOT_IMPLEMENTED"	},
+	{ OBEX_RSP_BAD_GATEWAY,			"BAD_GATEWAY"		},
+	{ OBEX_RSP_SERVICE_UNAVAILABLE,		"SERVICE_UNAVAILABLE"	},
+	{ OBEX_RSP_GATEWAY_TIMEOUT,		"GATEWAY_TIMEOUT"	},
+	{ OBEX_RSP_VERSION_NOT_SUPPORTED,	"VERSION_NOT_SUPPORTED"	},
+	{ OBEX_RSP_DATABASE_FULL,		"DATABASE_FULL"		},
+	{ OBEX_RSP_DATABASE_LOCKED,		"DATABASE_LOCKED"	},
+	{ 0xFF,					NULL			},
+};
+
+static void print_event(int evt, int cmd, int rsp)
+{
+	const char *evtstr = NULL, *cmdstr = NULL, *rspstr = NULL;
+	int i;
+	static int lastevt, lastcmd;
+
+	if (evt < 0)
+		evt = lastevt;
+	else
+		lastevt = evt;
+
+	if (cmd < 0)
+		cmd = lastcmd;
+	else
+		lastcmd = cmd;
+
+	for (i = 0; obex_event[i].evt != 0xFF; i++) {
+		if (obex_event[i].evt != evt)
+			continue;
+		evtstr = obex_event[i].name;
+	}
+
+	for (i = 0; obex_command[i].cmd != 0xFF; i++) {
+		if (obex_command[i].cmd != cmd)
+			continue;
+		cmdstr = obex_command[i].name;
+	}
+
+	for (i = 0; obex_response[i].rsp != 0xFF; i++) {
+		if (obex_response[i].rsp != rsp)
+			continue;
+		rspstr = obex_response[i].name;
+	}
+
+	obex_debug("%s(0x%x), %s(0x%x), %s(0x%x)", evtstr, evt, cmdstr, cmd,
+								rspstr, rsp);
+}
+
 static void os_set_response(obex_object_t *obj, int err)
 {
-	guint8 rsp;
-	guint8 lastrsp;
+	uint8_t rsp;
+	uint8_t lastrsp;
 
 	switch (err) {
 	case 0:
@@ -76,6 +228,7 @@ static void os_set_response(obex_object_t *obj, int err)
 		lastrsp = OBEX_RSP_SUCCESS;
 		break;
 	case -EPERM:
+	case -EACCES:
 		rsp = OBEX_RSP_FORBIDDEN;
 		lastrsp = OBEX_RSP_FORBIDDEN;
 		break;
@@ -95,22 +248,43 @@ static void os_set_response(obex_object_t *obj, int err)
 		rsp = OBEX_RSP_NOT_IMPLEMENTED;
 		lastrsp = OBEX_RSP_NOT_IMPLEMENTED;
 		break;
+	case -ENOTEMPTY:
+	case -EEXIST:
+		rsp = OBEX_RSP_PRECONDITION_FAILED;
+		lastrsp = OBEX_RSP_PRECONDITION_FAILED;
+		break;
 	default:
 		rsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
 		lastrsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
 	}
 
+	print_event(-1, -1, rsp);
+
 	OBEX_ObjectSetRsp(obj, rsp, lastrsp);
 }
 
+static void os_session_mark_aborted(struct obex_session *os)
+{
+	/* the session was already cancelled/aborted or size in unknown */
+	if (os->aborted || os->size == OBJECT_SIZE_UNKNOWN)
+		return;
+
+	os->aborted = (os->size != os->offset);
+}
+
 static void os_reset_session(struct obex_session *os)
 {
+	os_session_mark_aborted(os);
+	if (os->service && os->service->reset)
+		os->service->reset(os, os->service_data);
+
 	if (os->object) {
 		os->driver->set_io_watch(os->object, NULL, NULL);
 		os->driver->close(os->object);
 		os->object = NULL;
 		os->obj = NULL;
-		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path)
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path &&
+				os->driver->remove)
 			os->driver->remove(os->path);
 	}
 
@@ -126,30 +300,22 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+	if (os->path) {
+		g_free(os->path);
+		os->path = NULL;
+	}
 
 	os->driver = NULL;
 	os->aborted = FALSE;
+	os->pending = 0;
 	os->offset = 0;
 	os->size = OBJECT_SIZE_DELETE;
-	os->finished = 0;
-}
-
-static void os_session_mark_aborted(struct obex_session *os)
-{
-	/* the session was alredy cancelled/aborted */
-	if (os->aborted)
-		return;
-
-	os->aborted = os->size == OBJECT_SIZE_UNKNOWN ? FALSE :
-							os->size != os->offset;
 }
 
 static void obex_session_free(struct obex_session *os)
 {
 	sessions = g_slist_remove(sessions, os);
 
-	os_reset_session(os);
-
 	if (os->io)
 		g_io_channel_unref(os->io);
 
@@ -157,12 +323,12 @@ static void obex_session_free(struct obex_session *os)
 }
 
 /* From Imendio's GnomeVFS OBEX module (om-utils.c) */
-static time_t parse_iso8610(const gchar *val, int size)
+static time_t parse_iso8610(const char *val, int size)
 {
 	time_t time, tz_offset = 0;
 	struct tm tm;
-	gchar *date;
-	gchar tz;
+	char *date;
+	char tz;
 	int nr;
 
 	memset(&tm, 0, sizeof(tm));
@@ -203,35 +369,82 @@ static time_t parse_iso8610(const gchar *val, int size)
 	return time;
 }
 
+static uint8_t *extract_nonce(const uint8_t *buffer, unsigned int hlen)
+{
+	struct auth_header *hdr;
+	uint8_t *nonce = NULL;
+	uint32_t len = 0;
+
+	while (len < hlen) {
+		hdr = (void *) buffer + len;
+
+		switch (hdr->tag) {
+		case NONCE_TAG:
+			if (hdr->len != NONCE_LEN)
+				return NULL;
+
+			nonce = hdr->val;
+			break;
+		}
+
+		len += hdr->len + sizeof(struct auth_header);
+	}
+
+	return nonce;
+}
+
+static uint8_t *challenge_response(const uint8_t *nonce)
+{
+	GChecksum *md5;
+	uint8_t *result;
+	size_t size;
+
+	result = g_new0(uint8_t, NONCE_LEN);
+
+	md5 = g_checksum_new(G_CHECKSUM_MD5);
+	if (md5 == NULL)
+		return result;
+
+	g_checksum_update(md5, nonce, NONCE_LEN);
+	g_checksum_update(md5, (uint8_t *) ":BlueZ", 6);
+
+	size = NONCE_LEN;
+	g_checksum_get_digest(md5, result, &size);
+
+	g_checksum_free(md5);
+
+	return result;
+}
+
 static void cmd_connect(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
 	obex_connect_hdr_t *nonhdr;
 	obex_headerdata_t hd;
 	uint8_t *buffer;
-	guint hlen, newsize;
-	guint16 mtu;
-	guint8 hi;
-	const guint8 *target = NULL, *who = NULL;
-	guint target_size = 0, who_size = 0;
+	unsigned int hlen, newsize;
+	uint16_t mtu;
+	uint8_t hi;
+	const uint8_t *target = NULL, *who = NULL, *nonce = NULL;
+	unsigned int target_size = 0, who_size = 0;
 	int err;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		debug("Invalid OBEX CONNECT packet");
+		DBG("Invalid OBEX CONNECT packet");
 		return;
 	}
 
 	nonhdr = (obex_connect_hdr_t *) buffer;
 	mtu = g_ntohs(nonhdr->mtu);
-	debug("Version: 0x%02x. Flags: 0x%02x  OBEX packet length: %d",
+	DBG("Version: 0x%02x. Flags: 0x%02x  OBEX packet length: %d",
 			nonhdr->version, nonhdr->flags, mtu);
 	/* Leave space for headers */
 	newsize = mtu - 200;
 
 	os->tx_mtu = newsize;
 
-	debug("Resizing stream chunks to %d", newsize);
+	DBG("Resizing stream chunks to %d", newsize);
 
 	/* connection id will be used to track the sessions, even for OPP */
 	os->cid = ++cid;
@@ -246,6 +459,15 @@ static void cmd_connect(struct obex_session *os,
 			target = hd.bs;
 			target_size = hlen;
 			break;
+		case OBEX_HDR_AUTHCHAL:
+			if (nonce) {
+				DBG("Ignoring multiple challenge headers");
+				break;
+			}
+
+			nonce = extract_nonce(hd.bs, hlen);
+			DBG("AUTH CHALLENGE REQUEST");
+			break;
 		}
 	}
 
@@ -259,7 +481,7 @@ static void cmd_connect(struct obex_session *os,
 		return;
 	}
 
-	debug("Selected driver: %s", os->service->name);
+	DBG("Selected driver: %s", os->service->name);
 
 	if (!os->service->connect) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -278,15 +500,30 @@ static void cmd_connect(struct obex_session *os,
 				OBEX_FL_FIT_ONE_PACKET);
 	}
 
+	if (err == 0 && nonce) {
+		uint8_t challenge[18];
+		struct auth_header *hdr = (struct auth_header *) challenge;
+		uint8_t *response = challenge_response(nonce);
+
+		hdr->tag = DIGEST_TAG;
+		hdr->len = NONCE_LEN;
+		memcpy(hdr->val, response, NONCE_LEN);
+
+		g_free(response);
+
+		hd.bs = challenge;
+		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_AUTHRESP, hd, 18, 0);
+	}
+
 	os_set_response(obj, err);
 }
 
-static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
+static gboolean chk_cid(obex_t *obex, obex_object_t *obj, uint32_t cid)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	unsigned int hlen;
+	uint8_t hi;
 	gboolean ret = FALSE;
 
 	os = OBEX_GetUserData(obex);
@@ -311,11 +548,165 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 	return ret;
 }
 
+static int obex_read_stream(struct obex_session *os, obex_t *obex,
+						obex_object_t *obj)
+{
+	int size;
+	ssize_t len = 0;
+	const uint8_t *buffer;
+
+	DBG("name=%s type=%s rx_mtu=%d file=%p",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->rx_mtu, os->object);
+
+	if (os->aborted)
+		return -EPERM;
+
+	/* workaround: client didn't send the object lenght */
+	if (os->size == OBJECT_SIZE_DELETE)
+		os->size = OBJECT_SIZE_UNKNOWN;
+
+	/* If there's something to write and we are able to write it */
+	if (os->pending > 0 && os->driver)
+		goto write;
+
+	size = OBEX_ObjectReadStream(obex, obj, &buffer);
+	if (size < 0) {
+		error("Error on OBEX stream");
+		return -EIO;
+	}
+
+	if (size > os->rx_mtu) {
+		error("Received more data than RX_MAX");
+		return -EIO;
+	}
+
+	os->buf = g_realloc(os->buf, os->pending + size);
+	memcpy(os->buf + os->pending, buffer, size);
+	os->pending += size;
+
+	/* only write if both object and driver are valid */
+	if (os->object == NULL || os->driver == NULL) {
+		DBG("Stored %" PRIu64 " bytes into temporary buffer",
+								os->pending);
+		return 0;
+	}
+
+write:
+	while (os->pending > 0) {
+		ssize_t w;
+
+		w = os->driver->write(os->object, os->buf + len,
+					os->pending);
+		if (w < 0) {
+			if (w == -EINTR)
+				continue;
+			else {
+				memmove(os->buf, os->buf + len, os->pending);
+				return w;
+			}
+		}
+
+		len += w;
+		os->offset += w;
+		os->pending -= w;
+	}
+
+	/* Flush on EOS */
+	if (os->size != OBJECT_SIZE_UNKNOWN && os->size == os->offset &&
+							os->driver->flush)
+		return os->driver->flush(os->object) > 0 ? -EAGAIN : 0;
+
+	return 0;
+}
+
+static int obex_write_stream(struct obex_session *os,
+			obex_t *obex, obex_object_t *obj)
+{
+	obex_headerdata_t hd;
+	ssize_t len;
+	unsigned int flags;
+	uint8_t hi;
+
+	DBG("name=%s type=%s tx_mtu=%d file=%p",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->tx_mtu, os->object);
+
+	if (os->aborted)
+		return -EPERM;
+
+	if (os->object == NULL)
+		return -EIO;
+
+	len = os->driver->read(os->object, os->buf, os->tx_mtu, &hi);
+	if (len < 0) {
+		error("read(): %s (%zd)", strerror(-len), -len);
+		if (len == -EAGAIN)
+			return len;
+		else if (len == -ENOSTR)
+			return 0;
+
+		g_free(os->buf);
+		os->buf = NULL;
+		return len;
+	}
+
+	hd.bs = os->buf;
+
+	switch (hi) {
+	case OBEX_HDR_BODY:
+		flags = len ? OBEX_FL_STREAM_DATA : OBEX_FL_STREAM_DATAEND;
+		break;
+	case OBEX_HDR_APPARAM:
+		flags =  0;
+		break;
+	default:
+		error("read(): unkown header type %u", hi);
+		return -EIO;
+	}
+
+	OBEX_ObjectAddHeader(obex, obj, hi, hd, len, flags);
+
+	if (len == 0) {
+		g_free(os->buf);
+		os->buf = NULL;
+	}
+
+	return 0;
+}
+
+static gboolean handle_async_io(void *object, int flags, int err,
+						void *user_data)
+{
+	struct obex_session *os = user_data;
+	int ret = 0;
+
+	if (err < 0) {
+		ret = err;
+		goto proceed;
+	}
+
+	if (flags & (G_IO_IN | G_IO_PRI))
+		ret = obex_write_stream(os, os->obex, os->obj);
+	else if ((flags & G_IO_OUT) && os->pending > 0)
+		ret = obex_read_stream(os, os->obex, os->obj);
+
+proceed:
+	if (ret < 0) {
+		os_set_response(os->obj, err);
+		OBEX_CancelRequest(os->obex, TRUE);
+	} else
+		OBEX_ResumeRequest(os->obex);
+
+	return FALSE;
+}
+
 static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	gboolean stream;
+	unsigned int hlen;
+	uint8_t hi;
 	int err;
 
 	if (!os->service) {
@@ -333,20 +724,20 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 		switch (hi) {
 		case OBEX_HDR_NAME:
 			if (os->name) {
-				debug("Ignoring multiple name headers");
+				DBG("Ignoring multiple name headers");
 				break;
 			}
 
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
-			debug("OBEX_HDR_NAME: %s", os->name);
+			DBG("OBEX_HDR_NAME: %s", os->name);
 			break;
 		case OBEX_HDR_TYPE:
 			if (os->type) {
-				debug("Ignoring multiple type headers");
+				DBG("Ignoring multiple type headers");
 				break;
 			}
 
@@ -357,48 +748,74 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
-				debug("Invalid type header: %s", hd.bs);
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
+				DBG("Invalid type header: %s", hd.bs);
 				break;
 			}
 
 			/* FIXME: x-obex/folder-listing - type is mandatory */
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
-			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->type = g_strndup((const char *) hd.bs, hlen);
+			DBG("OBEX_HDR_TYPE: %s", os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target,
+						os->service->target_size,
+						os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 		}
 	}
 
+	if (os->type == NULL)
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							os->service->target_size,
+							NULL,
+							os->service->who,
+							os->service->who_size);
+
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
-		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+		error("No driver found");
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 					OBEX_RSP_NOT_IMPLEMENTED);
-			return;
-		}
+		return;
 	}
 
-	err = os->service->get(os, obj, os->service_data);
-	if (err == 0) {
-		if (os->size != OBJECT_SIZE_UNKNOWN) {
-			hd.bq4 = os->size;
-			OBEX_ObjectAddHeader(obex, obj,
-					OBEX_HDR_LENGTH, hd, 4, 0);
-		}
+	err = os->service->get(os, obj, &stream, os->service_data);
 
-		/* Add body header */
-		hd.bs = NULL;
-		if (os->size == 0)
-			OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
-					hd, 0, OBEX_FL_FIT_ONE_PACKET);
-		else
-			OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
-					hd, 0, OBEX_FL_STREAM_START);
+	if (err < 0)
+		goto done;
+
+	if (os->size != OBJECT_SIZE_UNKNOWN && os->size < UINT32_MAX) {
+		hd.bq4 = os->size;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_LENGTH, hd, 4, 0);
 	}
 
+	/* Add body header */
+	hd.bs = NULL;
+	if (os->size == 0) {
+		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
+						OBEX_FL_FIT_ONE_PACKET);
+		goto done;
+	}
+
+	if (stream)
+		/* Standard data stream */
+		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
+						OBEX_FL_STREAM_START);
+
+	/* Try to write to stream and suspend the stream immediately
+	 * if no data available to send. */
+	err = obex_write_stream(os, obex, obj);
+	if (err == -EAGAIN) {
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return;
+	}
+
+done:
 	os_set_response(obj, err);
 }
 
@@ -406,9 +823,9 @@ static void cmd_setpath(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint32 hlen;
+	uint32_t hlen;
 	int err;
-	guint8 hi;
+	uint8_t hi;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -431,7 +848,7 @@ static void cmd_setpath(struct obex_session *os,
 			continue;
 
 		if (os->name) {
-			debug("Ignoring multiple name headers");
+			DBG("Ignoring multiple name headers");
 			break;
 		}
 
@@ -442,10 +859,10 @@ static void cmd_setpath(struct obex_session *os,
 			break;
 		}
 
-		os->name = g_convert((const gchar *) hd.bs, hlen,
+		os->name = g_convert((const char *) hd.bs, hlen,
 				"UTF8", "UTF16BE", NULL, NULL, NULL);
 
-		debug("Set path name: %s", os->name);
+		DBG("Set path name: %s", os->name);
 		break;
 	}
 
@@ -453,20 +870,19 @@ static void cmd_setpath(struct obex_session *os,
 	os_set_response(obj, err);
 }
 
-int obex_stream_start(struct obex_session *os,
-		const gchar *filename, gpointer context)
+int obex_get_stream_start(struct obex_session *os, const char *filename)
 {
-	gint err;
-	gpointer object;
-	size_t size;
+	int err;
+	void *object;
+	size_t size = OBJECT_SIZE_UNKNOWN;
 
-	object = os->driver->open(filename, O_RDONLY, 0, context, &size, &err);
+	object = os->driver->open(filename, O_RDONLY, 0, os->service_data,
+								&size, &err);
 	if (object == NULL) {
 		error("open(%s): %s (%d)", filename, strerror(-err), -err);
-		goto fail;
+		return err;
 	}
 
-	os->driver->context = context;
 	os->object = object;
 	os->offset = 0;
 	os->size = size;
@@ -475,77 +891,14 @@ int obex_stream_start(struct obex_session *os,
 		os->buf = g_malloc0(os->tx_mtu);
 
 	return 0;
-
-fail:
-	if (object)
-		os->driver->close(object);
-
-	return err;
 }
 
-static gint obex_write_stream(struct obex_session *os,
-			obex_t *obex, obex_object_t *obj)
+int obex_put_stream_start(struct obex_session *os, const char *filename)
 {
-	obex_headerdata_t hd;
-	gint32 len;
-	guint8 *ptr;
-
-	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
-		os->name ? os->name : "", os->type ? os->type : "",
-		os->tx_mtu, os->object);
-
-	if (os->aborted)
-		return -EPERM;
-
-	if (os->object == NULL) {
-		if (os->buf == NULL && os->finished == FALSE)
-			return -EIO;
-
-		len = MIN(os->size - os->offset, os->tx_mtu);
-		ptr = os->buf + os->offset;
-		goto add_header;
-	}
-
-	len = os->driver->read(os->object, os->buf, os->tx_mtu);
-	if (len < 0) {
-		error("read(): %s (%d)", strerror(-len), -len);
-		if (len == -EAGAIN)
-			return len;
-
-		g_free(os->buf);
-		os->buf = NULL;
-		return len;
-	}
-
-	ptr = os->buf;
-
-add_header:
-
-	hd.bs = ptr;
-
-	if (len == 0) {
-		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
-					OBEX_FL_STREAM_DATAEND);
-		g_free(os->buf);
-		os->buf = NULL;
-		return len;
-	}
-
-	os->offset += len;
-
-	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
-				OBEX_FL_STREAM_DATA);
-
-	return len;
-}
-
-gint obex_prepare_put(struct obex_session *os, const gchar *filename)
-{
-	gint len;
 	int err;
 
 	os->object = os->driver->open(filename, O_WRONLY | O_CREAT | O_TRUNC,
-					0600, os->driver->context,
+					0600, os->service_data,
 					os->size != OBJECT_SIZE_UNKNOWN ?
 					(size_t *) &os->size : NULL, &err);
 	if (os->object == NULL) {
@@ -553,132 +906,26 @@ gint obex_prepare_put(struct obex_session *os, const gchar *filename)
 		return -EPERM;
 	}
 
-	os->path = filename;
+	os->path = g_strdup(filename);
 
 	if (!os->buf) {
-		debug("PUT request checked, no buffered data");
+		DBG("PUT request checked, no buffered data");
 		return 0;
 	}
 
-	len = 0;
-	while (len < os->offset) {
-		gint w;
-
-		w = os->driver->write(os->object, os->buf + len, os->offset - len);
-		if (w < 0) {
-			error("write(%s): %s (%d)", filename, strerror(-w), -w);
-			if (w == -EINTR)
-				continue;
-			else
-				return w;
-		}
-
-		len += w;
-	}
-
-	return 0;
-}
-
-static gint obex_read_stream(struct obex_session *os, obex_t *obex,
-				obex_object_t *obj)
-{
-	gint size;
-	gint32 len = 0;
-	const guint8 *buffer;
-
-	if (os->aborted)
-		return -EPERM;
-
-	/* workaround: client didn't send the object lenght */
-	if (os->size == OBJECT_SIZE_DELETE)
-		os->size = OBJECT_SIZE_UNKNOWN;
-
-	size = OBEX_ObjectReadStream(obex, obj, &buffer);
-	if (size < 0) {
-		error("Error on OBEX stream");
-		return -EIO;
-	}
-
-	if (size > os->rx_mtu) {
-		error("Received more data than RX_MAX");
-		return -EIO;
-	}
-
-	if (os->object == NULL && size > 0) {
-		os->buf = g_realloc(os->buf, os->offset + size);
-		memcpy(os->buf + os->offset, buffer, size);
-		os->offset += size;
-
-		debug("Stored %u bytes into temporary buffer", size);
-
+	if (os->pending == 0)
 		return 0;
-	}
 
-	while (len < size) {
-		gint w;
-
-		w = os->driver->write(os->object, buffer + len, size - len);
-		if (w < 0) {
-			if (w == -EINTR)
-				continue;
-			else
-				return w;
-		}
-
-		len += w;
-	}
-
-	os->offset += len;
-
-	return 0;
-}
-
-static gboolean handle_async_io(gpointer object, int flags, int err,
-				gpointer user_data)
-{
-	struct obex_session *os = user_data;
-	int ret = 0;
-
-	if (err < 0) {
-		ret = err;
-		goto proceed;
-	}
-
-	if (flags & (G_IO_IN | G_IO_PRI))
-		ret = obex_write_stream(os, os->obex, os->obj);
-	else if (flags & G_IO_OUT)
-		ret = obex_read_stream(os, os->obex, os->obj);
-
-proceed:
-	switch (ret) {
-	case -EINVAL:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		break;
-	case -EPERM:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
-					OBEX_RSP_FORBIDDEN);
-		break;
-	default:
-		if (ret < 0)
-			OBEX_ObjectSetRsp(os->obj,
-					OBEX_RSP_INTERNAL_SERVER_ERROR,
-					OBEX_RSP_INTERNAL_SERVER_ERROR);
-		break;
-	}
-
-	OBEX_ResumeRequest(os->obex);
-
-	return FALSE;
+	return obex_read_stream(os, os->obex, NULL);
 }
 
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
-	gint ret;
+	unsigned int hlen;
+	uint8_t hi;
+	int ret;
 
 	os = OBEX_GetUserData(obex);
 
@@ -696,21 +943,21 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		switch (hi) {
 		case OBEX_HDR_NAME:
 			if (os->name) {
-				debug("Ignoring multiple name headers");
+				DBG("Ignoring multiple name headers");
 				break;
 			}
 
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
-			debug("OBEX_HDR_NAME: %s", os->name);
+			DBG("OBEX_HDR_NAME: %s", os->name);
 			break;
 
 		case OBEX_HDR_TYPE:
 			if (os->type) {
-				debug("Ignoring multiple type headers");
+				DBG("Ignoring multiple type headers");
 				break;
 			}
 
@@ -721,14 +968,19 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
-				debug("Invalid type header: %s", hd.bs);
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
+				DBG("Invalid type header: %s", hd.bs);
 				break;
 			}
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
-			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->type = g_strndup((const char *) hd.bs, hlen);
+			DBG("OBEX_HDR_TYPE: %s", os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target,
+						os->service->target_size,
+						os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 
 		case OBEX_HDR_BODY:
@@ -738,35 +990,33 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 		case OBEX_HDR_LENGTH:
 			os->size = hd.bq4;
-			debug("OBEX_HDR_LENGTH: %d", os->size);
+			DBG("OBEX_HDR_LENGTH: %" PRIu64, os->size);
 			break;
 		case OBEX_HDR_TIME:
-			os->time = parse_iso8610((const gchar *) hd.bs, hlen);
+			os->time = parse_iso8610((const char *) hd.bs, hlen);
 			break;
 		}
 	}
 
 	OBEX_ObjectReParseHeaders(obex, obj);
 
+	if (os->type == NULL)
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							os->service->target_size,
+							NULL,
+							os->service->who,
+							os->service->who_size);
+
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
-		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+		error("No driver found");
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 					OBEX_RSP_NOT_IMPLEMENTED);
-			return FALSE;
-		}
+		return FALSE;
 	}
 
 	if (!os->service->chkput)
 		goto done;
 
-	if (!os->name) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		return FALSE;
-	}
-
 	ret = os->service->chkput(os, os->service_data);
 	switch (ret) {
 	case 0:
@@ -774,13 +1024,17 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	case -EBADR:
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
+					OBEX_RSP_BAD_REQUEST);
+		return FALSE;
 	case -EAGAIN:
 		OBEX_SuspendRequest(obex, obj);
 		os->obj = obj;
 		os->driver->set_io_watch(os->object, handle_async_io, os);
 		return TRUE;
 	default:
-		debug("Unhandled chkput error: %d", ret);
+		DBG("Unhandled chkput error: %d", ret);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
 				OBEX_RSP_INTERNAL_SERVER_ERROR);
 		return FALSE;
@@ -788,7 +1042,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	}
 
 	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
-		debug("Got a PUT without a Length");
+		DBG("Got a PUT without a Length");
 		goto done;
 	}
 
@@ -800,13 +1054,11 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
+	int err;
+
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
-	} else if (!os->service->put) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
-				OBEX_RSP_NOT_IMPLEMENTED);
-		return;
 	}
 
 	g_return_if_fail(chk_cid(obex, obj, os->cid));
@@ -816,27 +1068,51 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
-	os->service->put(os, os->service_data);
+	if (!os->service->put) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+				OBEX_RSP_NOT_IMPLEMENTED);
+		return;
+	}
+
+	err = os->service->put(os, obj, os->service_data);
+	if (err < 0) {
+		os_set_response(obj, err);
+		return;
+	}
+
+	/* Check if there is a body and it is not empty (size > 0), otherwise
+	   openobex won't notify us with OBEX_EV_STREAMAVAIL and it gonna reply
+	   right away */
+	if (os->size != 0)
+		return;
+
+	/* Flush immediatly since there is nothing to write so the driver
+	   has a chance to do something before we reply */
+	if (os->object && os->driver && os->driver->flush &&
+					os->driver->flush(os->object) > 0) {
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+	}
 }
 
-static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
-					gint evt, gint cmd, gint rsp)
+static void obex_event_cb(obex_t *obex, obex_object_t *obj, int mode,
+					int evt, int cmd, int rsp)
 {
 	struct obex_session *os;
+	int err;
 
-	obex_debug(evt, cmd, rsp);
+	print_event(evt, cmd, rsp);
 
 	os = OBEX_GetUserData(obex);
 
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
-		if (os->service->progress)
+		if (os->service && os->service->progress)
 			os->service->progress(os, os->service_data);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
-		if (os->service->reset)
-			os->service->reset(os, os->service_data);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
@@ -851,9 +1127,6 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		case OBEX_CMD_GET:
 		case OBEX_CMD_SETPATH:
 		default:
-			os_session_mark_aborted(os);
-			if (os->service->reset)
-				os->service->reset(os, os->service_data);
 			os_reset_session(os);
 			break;
 		}
@@ -904,48 +1177,32 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			cmd_put(os, obex, obj);
 			break;
 		default:
-			debug("Unknown request: 0x%X", cmd);
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
+			DBG("Unknown request: 0x%X", cmd);
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+						OBEX_RSP_NOT_IMPLEMENTED);
 			break;
 		}
 		break;
 	case OBEX_EV_STREAMAVAIL:
-		switch (obex_read_stream(os, obex, obj)) {
-		case 0:
-			break;
-		case -EPERM:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-			break;
-		case -EAGAIN:
-			OBEX_SuspendRequest(obex, obj);
-			os->obj = obj;
-			os->driver->set_io_watch(os->object, handle_async_io, os);
-			break;
-		default:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_INTERNAL_SERVER_ERROR,
-				OBEX_RSP_INTERNAL_SERVER_ERROR);
-			break;
-		}
-
-		break;
-	case OBEX_EV_STREAMEMPTY:
-		switch (obex_write_stream(os, obex, obj)) {
-		case -EPERM:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-			break;
-		case -EAGAIN:
+		err = obex_read_stream(os, obex, obj);
+		if (err == -EAGAIN) {
 			OBEX_SuspendRequest(obex, obj);
 			os->obj = obj;
 			os->driver->set_io_watch(os->object, handle_async_io,
-							os);
-			break;
-		default:
-			break;
-		}
+									os);
+		} else if (err < 0)
+			os_set_response(obj, err);
+
+		break;
+	case OBEX_EV_STREAMEMPTY:
+		err = obex_write_stream(os, obex, obj);
+		if (err == -EAGAIN) {
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io,
+									os);
+		} else if (err < 0)
+			os_set_response(obj, err);
 
 		break;
 	case OBEX_EV_LINKERR:
@@ -956,26 +1213,22 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		break;
 
 	default:
-		debug("Unknown evt %d", evt);
+		DBG("Unknown evt %d", evt);
 		break;
 	}
 }
 
-void server_free(struct server *server)
-{
-	g_free(server->folder);
-	g_free(server->capability);
-	g_free(server->devnode);
-	g_free(server);
-}
-
-static void obex_handle_destroy(gpointer user_data)
+static void obex_handle_destroy(void *user_data)
 {
 	struct obex_session *os;
 	obex_t *obex = user_data;
 
+	DBG("");
+
 	os = OBEX_GetUserData(obex);
 
+	os_reset_session(os);
+
 	if (os->service && os->service->disconnect)
 		os->service->disconnect(os, os->service_data);
 
@@ -984,88 +1237,44 @@ static void obex_handle_destroy(gpointer user_data)
 	OBEX_Cleanup(obex);
 }
 
-static gboolean tty_reinit(gpointer data)
-{
-	struct server *server = data;
-	GSList *l;
-	guint services = 0;
-
-	for (l = server->drivers; l; l = l->next) {
-		struct obex_service_driver *driver = l->data;
-
-		services |= driver->service;
-	}
-
-	tty_init(services, server->folder, server->capability,
-			server->symlinks, server->devnode);
-
-	server_free(server);
-
-	return FALSE;
-}
-
 static gboolean obex_handle_input(GIOChannel *io,
-				GIOCondition cond, gpointer user_data)
+				GIOCondition cond, void *user_data)
 {
 	obex_t *obex = user_data;
-	struct obex_session *os = OBEX_GetUserData(obex);
 
 	if (cond & (G_IO_HUP | G_IO_ERR | G_IO_NVAL)) {
 		error("obex_handle_input: poll event %s%s%s",
 				(cond & G_IO_HUP) ? "HUP " : "",
 				(cond & G_IO_ERR) ? "ERR " : "",
 				(cond & G_IO_NVAL) ? "NVAL " : "");
-		goto failed;
+		return FALSE;
 	}
 
 	if (OBEX_HandleInput(obex, 1) < 0) {
 		error("Handle input error");
-		goto failed;
+		return FALSE;
 	}
 
 	return TRUE;
-
-failed:
-	if (os->server->devnode) {
-		if (cond & G_IO_NVAL)
-			tty_closed();
-		else
-			g_idle_add(tty_reinit, os->server);
-	}
-
-	return FALSE;
 }
 
-void obex_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
-{
-	struct server *server = user_data;
-
-	if (err) {
-		error("%s", err->message);
-		g_io_channel_shutdown(io, TRUE, NULL);
-		return;
-	}
-
-	if (obex_session_start(io, server) < 0)
-		g_io_channel_shutdown(io, TRUE, NULL);
-}
-
-gint obex_session_start(GIOChannel *io, struct server *server)
+int obex_session_start(GIOChannel *io, uint16_t tx_mtu, uint16_t rx_mtu,
+			struct obex_server *server)
 {
 	struct obex_session *os;
 	obex_t *obex;
-	gint ret, fd;
+	int ret, fd;
 
 	os = g_new0(struct obex_session, 1);
 
-	os->service = obex_service_driver_find(server->drivers, NULL, 0,
-						NULL, 0);
+	os->service = obex_service_driver_find(server->drivers, NULL,
+							0, NULL, 0);
 	os->server = server;
-	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
-	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
+	os->rx_mtu = rx_mtu != 0 ? rx_mtu : DEFAULT_RX_MTU;
+	os->tx_mtu = tx_mtu != 0 ? tx_mtu : DEFAULT_TX_MTU;
 	os->size = OBJECT_SIZE_DELETE;
 
-	obex = OBEX_Init(OBEX_TRANS_FD, obex_event, 0);
+	obex = OBEX_Init(OBEX_TRANS_FD, obex_event_cb, 0);
 	if (!obex) {
 		obex_session_free(os);
 		return -EIO;
@@ -1095,25 +1304,18 @@ gint obex_session_start(GIOChannel *io, struct server *server)
 	return 0;
 }
 
-gint obex_tty_session_stop(void)
-{
-	GSList *l;
-
-	for (l = sessions; l != NULL; l = l->next) {
-		struct obex_session *os = l->data;
-
-		if (os->server->devnode && os->io)
-			g_io_channel_shutdown(os->io, TRUE, NULL);
-	}
-
-	return 0;
-}
-
 const char *obex_get_name(struct obex_session *os)
 {
 	return os->name;
 }
 
+void obex_set_name(struct obex_session *os, const char *name)
+{
+	g_free(os->name);
+	os->name = g_strdup(name);
+	DBG("Name changed: %s", os->name);
+}
+
 ssize_t obex_get_size(struct obex_session *os)
 {
 	return os->size;
@@ -1129,7 +1331,7 @@ const char *obex_get_root_folder(struct obex_session *os)
 	return os->server->folder;
 }
 
-guint16 obex_get_service(struct obex_session *os)
+uint16_t obex_get_service(struct obex_session *os)
 {
 	return os->service->service;
 }
@@ -1161,8 +1363,8 @@ int obex_remove(struct obex_session *os, const char *path)
 char *obex_get_id(struct obex_session *os)
 {
 	GError *gerr = NULL;
-	gchar address[18];
-	guint8 channel;
+	char address[18];
+	uint8_t channel;
 
 	bt_io_get(os->io, BT_IO_RFCOMM, &gerr,
 			BT_IO_OPT_DEST, address,
@@ -1175,11 +1377,11 @@ char *obex_get_id(struct obex_session *os)
 }
 
 ssize_t obex_aparam_read(struct obex_session *os,
-		obex_object_t *obj, const guint8 **buffer)
+		obex_object_t *obj, const uint8_t **buffer)
 {
 	obex_headerdata_t hd;
-	guint8 hi;
-	guint32 hlen;
+	uint8_t hi;
+	uint32_t hlen;
 
 	OBEX_ObjectReParseHeaders(os->obex, obj);
 
@@ -1192,3 +1394,28 @@ ssize_t obex_aparam_read(struct obex_session *os,
 
 	return -EBADR;
 }
+
+int obex_aparam_write(struct obex_session *os,
+		obex_object_t *obj, const uint8_t *data, unsigned int size)
+{
+	obex_headerdata_t hd;
+
+	hd.bs = data;
+
+	return OBEX_ObjectAddHeader(os->obex, obj,
+			OBEX_HDR_APPARAM, hd, size, 0);
+}
+
+int memncmp0(const void *a, size_t na, const void *b, size_t nb)
+{
+	if (na != nb)
+		return na - nb;
+
+	if (a == NULL)
+		return -(a != b);
+
+	if (b == NULL)
+		return a != b;
+
+	return memcmp(a, b, na);
+}
